#include <signal.h>																																																																																						
#include <vector>
#include <string>
#include <sys/stat.h>
#include <ros/ros.h>
#include <ros/package.h>
#include <sensor_msgs/PointCloud2.h>

// PCL specific includes
#include <pcl/conversions.h>
#include <pcl_conversions/pcl_conversions.h>
#include <pcl/point_cloud.h>
#include <pcl/console/parse.h>
#include <pcl/point_types.h>
#include <pcl/io/openni_grabber.h>
#include <pcl/sample_consensus/sac_model_plane.h>
#include <pcl/common/time.h>
#include <pcl/common/common.h>
#include <pcl/filters/crop_box.h>
#include <pcl/filters/passthrough.h>
#include <pcl/filters/voxel_grid.h>
#include <pcl/filters/extract_indices.h>
#include <pcl/ModelCoefficients.h>
#include <pcl/sample_consensus/method_types.h>
#include <pcl/sample_consensus/model_types.h>
#include <pcl/segmentation/sac_segmentation.h>
#include <pcl/segmentation/extract_clusters.h>
#include <pcl/kdtree/kdtree.h>

/* define what kind of point clouds we're using */
typedef pcl::PointXYZRGB PointT;
typedef pcl::PointCloud<PointT> PointCloudT;

//some global variables to store data
PointCloudT::Ptr cloud_in (new PointCloudT);
PointCloudT::Ptr cloud_out (new PointCloudT);
PointCloudT::Ptr cloud_plane (new PointCloudT);
PointCloudT::Ptr cloud_blobs (new PointCloudT);
sensor_msgs::PointCloud2 cloud_ros;


//true if Ctrl-C is pressed
bool g_caught_sigint=false;

//publisher for output
ros::Publisher cloud_pub;
ros::Publisher cloud_pub_blobs;

#define NUM_COLORS 6
int colors[NUM_COLORS][3] = { {255,255,255},
					 {0,255,0},
					 {0,0,255},
					 {255,0,255},
					 {255,255,0},
					 {0,255,255} };




/* what happens when ctr-c is pressed */
void sig_handler(int sig)
{
  g_caught_sigint = true;
  ROS_INFO("caught sigint, init shutdown sequence...");
  ros::shutdown();
  exit(1);
};

void colorPointCloud(PointCloudT::Ptr in, int color[3]){
	for (unsigned int i = 0; i < in->points.size(); i++){
		in->points.at(i).r = color[0];
		in->points.at(i).g = color[1];
		in->points.at(i).b = color[2];
	}
}

void
cloud_cb (const sensor_msgs::PointCloud2ConstPtr& input)
{

	//convert to PCL format
	pcl::fromROSMsg (*input, *cloud_in);

	// Apply z filter -- we don't care for anything X m away in the z direction
	pcl::PassThrough<PointT> pass;
	pass.setInputCloud (cloud_in);
	pass.setFilterFieldName ("z");
	pass.setFilterLimits (0.0, 1.35);
	pass.filter (*cloud_out);
	
	//Apply voxel grid filter
	pcl::VoxelGrid<PointT> vg;
	vg.setInputCloud (cloud_out);
	vg.setLeafSize (0.005f, 0.005f, 0.005f);
	vg.filter (*cloud_out);
	
	//Fit a plane to the point cloud
	
	//here, we will store the resulting plane coefficients
	pcl::ModelCoefficients::Ptr coefficients (new pcl::ModelCoefficients ());
	
	//here, we wil store the indices of the points that match the plane
	pcl::PointIndices::Ptr inliers (new pcl::PointIndices ());
	
	// Create the segmentation object
	pcl::SACSegmentation<PointT> seg;
	
	// Optional
	seg.setOptimizeCoefficients (true);
	
	// Mandatory
	seg.setModelType (pcl::SACMODEL_PLANE);
	seg.setMethodType (pcl::SAC_RANSAC);
	seg.setMaxIterations (1000);
	seg.setDistanceThreshold (0.03);

	// Create the filtering object
	pcl::ExtractIndices<PointT> extract;
	
	// Segment the largest planar component from the remaining cloud
	seg.setInputCloud (cloud_out);
	seg.segment (*inliers, *coefficients);

	// Extract the plane
	extract.setInputCloud (cloud_out);
	extract.setIndices (inliers);
	extract.setNegative (false);
	extract.filter (*cloud_plane);
	
	// Set the plane cloud's pixel to red
	for (unsigned int i = 0; i < cloud_plane->points.size(); i++){
		cloud_plane->points.at(i).r = 255;
		cloud_plane->points.at(i).g = 0;
		cloud_plane->points.at(i).b = 0;
	}
	
	//extract everything else
	extract.setNegative (true);
	extract.filter (*cloud_blobs);
	
	//Euclidean cluster extraction
	
	//create the KdTree structure for the input cloud
	pcl::search::KdTree<PointT>::Ptr tree (new pcl::search::KdTree<PointT>);
	tree->setInputCloud (cloud_blobs);

	//create the Cluster Extraction object
	std::vector<pcl::PointIndices> cluster_indices;
	pcl::EuclideanClusterExtraction<PointT> ec;
	ec.setClusterTolerance (0.02); // 2cm
	ec.setMinClusterSize (50);
	ec.setMaxClusterSize (25000);
	ec.setSearchMethod (tree);
	ec.setInputCloud (cloud_blobs);
	ec.extract (cluster_indices);

	//put the resulting clusters in a vector of point clouds
	std::vector<PointCloudT::Ptr > clusters;
	
	int cluster_counter = 0;
	
	//for each cluster...
	for (std::vector<pcl::PointIndices>::const_iterator it = cluster_indices.begin (); it != cluster_indices.end (); ++it)
	{
		PointCloudT::Ptr cloud_cluster (new PointCloudT);
		
		//for each point in the cluster
		for (std::vector<int>::const_iterator pit = it->indices.begin (); pit != it->indices.end (); pit++)
			cloud_cluster->points.push_back (cloud_blobs->points[*pit]); 
		
		cloud_cluster->width = cloud_cluster->points.size ();
		cloud_cluster->height = 1;
		cloud_cluster->is_dense = true;

		//color the cluster
		if (cluster_counter < NUM_COLORS-1){
			colorPointCloud(cloud_cluster,colors[cluster_counter]);
		}
		else {
			colorPointCloud(cloud_cluster,colors[NUM_COLORS-1]);
		}
		cluster_counter++;
		

		clusters.push_back(cloud_cluster);
	}
	
	ROS_INFO("[pcl_example_plane_objects.cpp] Found %i clusters!",(int)clusters.size());
	
	//publish plane point cloud
	pcl::toROSMsg(*cloud_plane,cloud_ros);
	cloud_ros.header.frame_id = cloud_in->header.frame_id;
	cloud_pub.publish(cloud_ros);
	
	//publish cloud containing all clusters
	cloud_blobs->clear();

	//"add" all clusters into a single point cloud
	for (unsigned int i = 0; i < clusters.size(); i++){
		*cloud_blobs += *clusters.at(i);
	}
	
	pcl::toROSMsg(*cloud_blobs,cloud_ros);
	cloud_ros.header.frame_id = cloud_in->header.frame_id;
	cloud_pub_blobs.publish(cloud_ros);
}

int main (int argc, char** argv)
{
	// Initialize ROS
	ros::init (argc, argv, "pcl_example");
	ros::NodeHandle nh;

	// Create a ROS subscriber for the input point cloud
	std::string param_topic = "/camera/depth_registered/points";
	ros::Subscriber sub = nh.subscribe (param_topic, 1, cloud_cb);

	//debugging publisher
	cloud_pub = nh.advertise<sensor_msgs::PointCloud2>("pcl_example/cloud_out", 1);
	cloud_pub_blobs = nh.advertise<sensor_msgs::PointCloud2>("pcl_example/cloud_blobs", 1);

	//register ctrl-c
	signal(SIGINT, sig_handler);

	//refresh rate
	double ros_rate = 5.0;
	ros::Rate r(ros_rate);

	// Main loop:
	while (!g_caught_sigint && ros::ok())
	{
		//collect messages
		ros::spinOnce();

		r.sleep();

	}
}
